#include <dds/DCPS/ConfigStoreImpl.h>

#include <dds/DCPS/Qos_Helper.h>

#include <gtestWrapper.h>

using namespace OpenDDS::DCPS;

TEST(dds_DCPS_ConfigPair, ctor)
{
  {
    ConfigPair cp("", "value");
    EXPECT_EQ(cp.key(), "");
    EXPECT_EQ(cp.value(), "value");
  }

  {
    ConfigPair cp("^", "value");
    EXPECT_EQ(cp.key(), "");
    EXPECT_EQ(cp.value(), "value");
  }

  {
    ConfigPair cp("^&*", "value");
    EXPECT_EQ(cp.key(), "");
    EXPECT_EQ(cp.value(), "value");
  }

  {
    ConfigPair cp("~!abc.123__CamelCase/CAMELCase#$%", "value");
    EXPECT_EQ(cp.key(), "ABC_123_CAMEL_CASE_CAMEL_CASE");
    EXPECT_EQ(cp.value(), "value");
  }

  {
    ConfigPair cp("CamelCase", "value");
    EXPECT_EQ(cp.key(), "CAMEL_CASE");
    EXPECT_EQ(cp.value(), "value");
  }

  {
    ConfigPair cp("##CamelCase##", "value");
    EXPECT_EQ(cp.key(), "CAMEL_CASE");
    EXPECT_EQ(cp.value(), "value");
  }
}

TEST(dds_DCPS_ConfigPair, key_has_prefix)
{
  ConfigPair cp("key", "value");
  EXPECT_TRUE(cp.key_has_prefix(""));
  EXPECT_TRUE(cp.key_has_prefix("k"));
  EXPECT_TRUE(cp.key_has_prefix("ke"));
  EXPECT_TRUE(cp.key_has_prefix("key"));
  EXPECT_FALSE(cp.key_has_prefix("keya"));
  EXPECT_FALSE(cp.key_has_prefix("a"));
}

TEST(dds_DCPS_ConfigStoreImpl, has)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store(topic);
  EXPECT_FALSE(store.has("key"));
  store.set_boolean("key", true);
  EXPECT_TRUE(store.has("key"));
}

TEST(dds_DCPS_ConfigStoreImpl, set_get_boolean)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store(topic);
  EXPECT_TRUE(store.get_boolean("key", true));
  EXPECT_FALSE(store.get_boolean("key", false));
  store.set_boolean("key", "true");
  EXPECT_TRUE(store.get_boolean("key", false));
  store.set_string("key", "not a boolean");
  EXPECT_TRUE(store.get_boolean("key", true));
}

TEST(dds_DCPS_ConfigStoreImpl, set_get_int32)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store(topic);
  EXPECT_EQ(store.get_int32("key", -37), -37);
  store.set_int32("key", -38);
  EXPECT_EQ(store.get_int32("key", -37), -38);
  store.set_string("key", "not an int32");
  EXPECT_EQ(store.get_int32("key", -37), -37);
}

TEST(dds_DCPS_ConfigStoreImpl, set_get_uint32)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store(topic);
  EXPECT_EQ(store.get_uint32("key", 37), 37u);
  store.set_uint32("key", 38);
  EXPECT_EQ(store.get_uint32("key", 37), 38u);
  store.set_string("key", "not a uint32");
  EXPECT_EQ(store.get_uint32("key", 37), 37u);
}

TEST(dds_DCPS_ConfigStoreImpl, set_get_float64)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store(topic);
  EXPECT_EQ(store.get_float64("key", 1.5), 1.5);
  store.set_float64("key", 2);
  EXPECT_EQ(store.get_float64("key", 1.5), 2);
  store.set_string("key", "not a float64");
  EXPECT_EQ(store.get_float64("key", 1.5), 1.5);
}

TEST(dds_DCPS_ConfigStoreImpl, set_get_string)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store(topic);
  CORBA::String_var str = store.get_string("key", "default");
  EXPECT_STREQ(str, "default");
  store.set_string("key", "not default");
  str = store.get_string("key", "default");
  EXPECT_STREQ(str, "not default");
}

TEST(dds_DCPS_ConfigStoreImpl, set_get_duration)
{
  using OpenDDS::DCPS::operator==;

  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store(topic);
  const DDS::Duration_t default_duration = make_duration_t(1,2);
  const DDS::Duration_t duration = make_duration_t(3,4);
  EXPECT_TRUE(store.get_duration("key", default_duration) == default_duration);
  store.set_duration("key", duration);
  EXPECT_TRUE(store.get_duration("key", default_duration) == duration);
  store.set_string("key", "not a duration");
  EXPECT_TRUE(store.get_duration("key", default_duration) == default_duration);
}

TEST(dds_DCPS_ConfigStoreImpl, unset)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store(topic);
  store.set_boolean("key", true);
  store.unset("key");
  EXPECT_FALSE(store.has("key"));
}

TEST(dds_DCPS_ConfigStoreImpl, set_get_String)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store(topic);
  const String default_string = "default";
  const String other_string = "other";
  EXPECT_EQ(store.get("key", default_string), default_string);
  store.set("key", other_string);
  EXPECT_EQ(store.get("key", default_string), other_string);

  store.set("key", "");
  EXPECT_EQ(store.get("key", default_string, false), default_string);
}

TEST(dds_DCPS_ConfigStoreImpl, set_get_TimeDuration_seconds)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store(topic);
  const TimeDuration default_duration(1,2);
  const TimeDuration duration(3);
  EXPECT_EQ(store.get("key", default_duration, ConfigStoreImpl::Format_IntegerSeconds), default_duration);
  store.set("key", duration, ConfigStoreImpl::Format_IntegerSeconds);
  EXPECT_EQ(store.get("key", default_duration, ConfigStoreImpl::Format_IntegerSeconds), duration);
  store.set_string("key", "not a duration");
  EXPECT_EQ(store.get("key", default_duration, ConfigStoreImpl::Format_IntegerSeconds), default_duration);
}

TEST(dds_DCPS_ConfigStoreImpl, set_get_TimeDuration_milliseconds)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store(topic);
  const TimeDuration default_duration(1,2);
  const TimeDuration duration(3,4000);
  EXPECT_EQ(store.get("key", default_duration, ConfigStoreImpl::Format_IntegerMilliseconds), default_duration);
  store.set("key", duration, ConfigStoreImpl::Format_IntegerMilliseconds);
  EXPECT_EQ(store.get("key", default_duration, ConfigStoreImpl::Format_IntegerMilliseconds), duration);
  store.set_string("key", "not a duration");
  EXPECT_EQ(store.get("key", default_duration, ConfigStoreImpl::Format_IntegerMilliseconds), default_duration);
}

TEST(dds_DCPS_ConfigStoreImpl, get_NetworkAddress)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store(topic);

  {
    const NetworkAddress default_value("0.0.0.0:0");
    EXPECT_EQ(store.get("key", default_value, ConfigStoreImpl::Format_No_Port, ConfigStoreImpl::Kind_IPV4), default_value);

    {
      const NetworkAddress value_no_port("127.0.0.1:0");
      store.set("key", value_no_port, ConfigStoreImpl::Format_No_Port, ConfigStoreImpl::Kind_IPV4);
      EXPECT_EQ(store.get("key", default_value, ConfigStoreImpl::Format_No_Port, ConfigStoreImpl::Kind_IPV4), value_no_port);
    }

    {
      const NetworkAddress value_required_port("127.0.0.1:80");
      store.set("key", value_required_port, ConfigStoreImpl::Format_Required_Port, ConfigStoreImpl::Kind_IPV4);
      EXPECT_EQ(store.get("key", default_value, ConfigStoreImpl::Format_Required_Port, ConfigStoreImpl::Kind_IPV4), value_required_port);
    }

    {
      const NetworkAddress value_optional_port("127.0.0.1:0");
      store.set("key", value_optional_port, ConfigStoreImpl::Format_Optional_Port, ConfigStoreImpl::Kind_IPV4);
      EXPECT_EQ(store.get("key", default_value, ConfigStoreImpl::Format_Optional_Port, ConfigStoreImpl::Kind_IPV4), value_optional_port);
    }

    {
      const NetworkAddress value_optional_port("127.0.0.1:80");
      store.set("key", value_optional_port, ConfigStoreImpl::Format_Optional_Port, ConfigStoreImpl::Kind_IPV4);
      EXPECT_EQ(store.get("key", default_value, ConfigStoreImpl::Format_Optional_Port, ConfigStoreImpl::Kind_IPV4), value_optional_port);
    }

    {
      store.set_string("key", "not a network address");
      EXPECT_EQ(store.get("key", default_value, ConfigStoreImpl::Format_Optional_Port, ConfigStoreImpl::Kind_IPV4), default_value);
    }

    {
      const NetworkAddress value_required_port("127.0.0.1:80");
      store.set("key", value_required_port, ConfigStoreImpl::Format_Required_Port, ConfigStoreImpl::Kind_ANY);
      EXPECT_EQ(store.get("key", default_value, ConfigStoreImpl::Format_Required_Port, ConfigStoreImpl::Kind_ANY), value_required_port);
    }
  }

#if ACE_HAS_IPV6
  {
    const NetworkAddress default_value("::");
    EXPECT_EQ(store.get("key6", default_value, ConfigStoreImpl::Format_No_Port, ConfigStoreImpl::Kind_IPV6), default_value);

    {
      const NetworkAddress value_no_port("::1");
      store.set("key6", value_no_port, ConfigStoreImpl::Format_No_Port, ConfigStoreImpl::Kind_IPV6);
      EXPECT_EQ(store.get("key6", default_value, ConfigStoreImpl::Format_No_Port, ConfigStoreImpl::Kind_IPV6), value_no_port);
    }

    {
      const NetworkAddress value_required_port("[::1]:80");
      store.set("key6", value_required_port, ConfigStoreImpl::Format_Required_Port, ConfigStoreImpl::Kind_IPV6);
      EXPECT_EQ(store.get("key6", default_value, ConfigStoreImpl::Format_Required_Port, ConfigStoreImpl::Kind_IPV6), value_required_port);
    }

    {
      const NetworkAddress value_optional_port("::1");
      store.set("key6", value_optional_port, ConfigStoreImpl::Format_Optional_Port, ConfigStoreImpl::Kind_IPV6);
      EXPECT_EQ(store.get("key6", default_value, ConfigStoreImpl::Format_Optional_Port, ConfigStoreImpl::Kind_IPV6), value_optional_port);
    }

    {
      const NetworkAddress value_optional_port("[::1]:80");
      store.set("key6", value_optional_port, ConfigStoreImpl::Format_Optional_Port, ConfigStoreImpl::Kind_IPV6);
      EXPECT_EQ(store.get("key6", default_value, ConfigStoreImpl::Format_Optional_Port, ConfigStoreImpl::Kind_IPV6), value_optional_port);
    }

    {
      store.set_string("key6", "not a network address");
      EXPECT_EQ(store.get("key6", default_value, ConfigStoreImpl::Format_Optional_Port, ConfigStoreImpl::Kind_IPV6), default_value);
    }

    {
      const NetworkAddress value_required_port("[::1]:80");
      store.set("key6", value_required_port, ConfigStoreImpl::Format_Required_Port, ConfigStoreImpl::Kind_ANY);
      EXPECT_EQ(store.get("key6", default_value, ConfigStoreImpl::Format_Required_Port, ConfigStoreImpl::Kind_ANY), value_required_port);
    }
  }
#endif
}

TEST(dds_DCPS_ConfigStoreImpl, take_has_prefix)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl store1(topic);
  ConfigReader_rch reader = make_rch<ConfigReader>(store1.datareader_qos());
  topic->connect(reader);

  store1.set("PREFIX1_key", "value");
  EXPECT_TRUE(take_has_prefix(reader, "PREFIX1"));
  EXPECT_FALSE(take_has_prefix(reader, "PREFIX1"));

  topic->disconnect(reader);
}

namespace {
  class Listener : public ConfigListener {
  public:
    Listener(JobQueue_rch job_queue)
      : ConfigListener(job_queue)
    {}

    MOCK_METHOD1(on_data_available, void(ConfigReader_rch));
  };
}

TEST(dds_DCPS_ConfigStoreImpl, process_section)
{
  JobQueue_rch job_queue = make_rch<JobQueue>(ACE_Reactor::instance());
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl config_store(topic);
  config_store.set("MYPREFIX_MY_SECTION_THIRDKEY", "firstvalue");
  ConfigReader_rch reader = make_rch<ConfigReader>(config_store.datareader_qos());
  RcHandle<Listener> listener = make_rch<Listener>(job_queue);
  ACE_Configuration_Heap config;
  ACE_Configuration_Section_Key section_key;
  config.open();
  config.open_section(config.root_section(), ACE_TEXT("my_section"), true, section_key);
  config.set_string_value(section_key, ACE_TEXT("mykey"), ACE_TEXT("myvalue"));
  config.set_string_value(section_key, ACE_TEXT("anotherkey"), ACE_TEXT("$file"));
  config.set_string_value(section_key, ACE_TEXT("thirdkey"), ACE_TEXT("secondvalue"));

  EXPECT_CALL(*listener.get(), on_data_available(reader)).Times(2);

  process_section(config_store, reader, listener, "MYPREFIX", config, config.root_section(), "my file name", false);

  EXPECT_EQ(config_store.get("MYPREFIX_MY_SECTION_MYKEY", "default"), "myvalue");
  EXPECT_EQ(config_store.get("MYPREFIX_MY_SECTION_ANOTHERKEY", "default"), "my file name");
  EXPECT_EQ(config_store.get("MYPREFIX_MY_SECTION_THIRDKEY", "default"), "firstvalue");
}

TEST(dds_DCPS_ConfigStoreImpl, process_section_allow_overwrite)
{
  ConfigTopic_rch topic = make_rch<ConfigTopic>();
  ConfigStoreImpl config_store(topic);
  config_store.set("MYPREFIX_MY_SECTION_THIRDKEY", "firstvalue");
  ACE_Configuration_Heap config;
  ACE_Configuration_Section_Key section_key;
  config.open();
  config.open_section(config.root_section(), ACE_TEXT("my_section"), true, section_key);
  config.set_string_value(section_key, ACE_TEXT("mykey"), ACE_TEXT("myvalue"));
  config.set_string_value(section_key, ACE_TEXT("anotherkey"), ACE_TEXT("$file"));
  config.set_string_value(section_key, ACE_TEXT("thirdkey"), ACE_TEXT("secondvalue"));

  process_section(config_store, ConfigReader_rch(), ConfigReaderListener_rch(), "MYPREFIX", config, config.root_section(), "my file name", true);

  EXPECT_EQ(config_store.get("MYPREFIX_MY_SECTION_MYKEY", "default"), "myvalue");
  EXPECT_EQ(config_store.get("MYPREFIX_MY_SECTION_ANOTHERKEY", "default"), "my file name");
  EXPECT_EQ(config_store.get("MYPREFIX_MY_SECTION_THIRDKEY", "default"), "secondvalue");
}
